package com.qmth.teachcloud.mark.utils;

import com.alibaba.fastjson.JSON;
import com.qmth.teachcloud.common.util.DateDisposeUtils;
import org.apache.commons.collections4.CollectionUtils;

import java.util.*;

/**
 * 链表模式实现的任务互斥锁工具
 */
public class TaskLock {

    // 固定头节点
    private LockNode head;

    // 总量计数
    private int count;

    public TaskLock() {
        head = new LockNode(0, 0, new HashSet<>());
        count = 0;
    }

    public synchronized boolean add(Object id, Object owner, Set<Long> questions) {
        if (head.next == null) {
            head.append(id, owner, questions);
            count++;
            return true;
        } else {
            LockNode node = head.next;
            while (true) {
                // id+questions，只能被一个owner获取
                if (node.isId(id) && node.isEqualQuestions(questions)) {
                    return false;
                }
                // id只能被一个owner领取一个number
                else if (node.isId(id) && !node.isOwner(owner)) {
                    return false;
                }
                // 跳到下一个node
                else if (node.hasNext()) {
                    node = node.next;
                }
                // 可以领取
                else {
                    node.append(id, owner, questions);
                    count++;
                    return true;
                }
            }
        }
    }

    public synchronized boolean exist(Object id, Object owner) {
        LockNode node = head.next;
        while (node != null) {
            if (node.isId(id)) {
                return node.isOwner(owner);
            } else {
                node = node.next;
            }
        }
        return false;
    }

    public synchronized boolean remove(Object id, Object owner) {
        LockNode node = head.next;
        while (node != null) {
            if (node.isId(id)) {
                if (node.isOwner(owner)) {
                    node.remove();
                    count--;
                    return true;
                } else {
                    return false;
                }
            } else {
                node = node.next;
            }
        }
        return false;
    }

    public synchronized boolean remove(Object id) {
        LockNode node = head.next;
        while (node != null) {
            if (node.isId(id)) {
                node.remove();
                count--;
                return true;
            } else {
                node = node.next;
            }
        }
        return false;
    }

    public synchronized void clear() {
        head.next = null;
        count = 0;
    }

    public synchronized void clear(Object owner) {
        LockNode node = head.next;
        while (node != null) {
            if (node.isOwner(owner)) {
                node.remove();
                count--;
            }
            node = node.next;
        }
    }

    public synchronized void expire(long expireTime) {
        long current = System.currentTimeMillis();
        LockNode node = head.next;
        while (node != null) {
            if ((current - node.time) > expireTime) {
                node.remove();
                count--;
            }
            node = node.next;
        }
    }

    public synchronized boolean refresh(Object id, Object owner) {
        LockNode node = head.next;
        while (node != null) {
            if (node.isId(id) && node.isOwner(owner)) {
                node.time = System.currentTimeMillis();
                return true;
            }
            node = node.next;
        }
        return false;
    }

    public synchronized void refresh(Object owner) {
        LockNode node = head.next;
        while (node != null) {
            if (node.isOwner(owner)) {
                node.time = System.currentTimeMillis();
            }
            node = node.next;
        }
    }

    public int count() {
        return count;
    }

    public int count(Object owner) {
        int count = 0;
        LockNode node = head.next;
        while (node != null) {
            if (node.isOwner(owner)) {
                count++;
            }
            node = node.next;
        }
        return count;
    }

    public int count(Long questionId) {
        int count = 0;
        LockNode node = head.next;
        while (node != null) {
            if (node.containsQuestionId(questionId)) {
                count++;
            }
            node = node.next;
        }
        return count;
    }

    public int count(Long questionId, Long userId) {
        int count = 0;
        LockNode node = head.next;
        while (node != null) {
            if (node.isOwner(userId) && node.containsQuestionId(questionId)) {
                count++;
            }
            node = node.next;
        }
        return count;
    }

    public List<Map<String, Object>> list() {
        List<Map<String, Object>> list = new LinkedList<>();
        LockNode node = head.next;
        while (node != null) {
            Map<String, Object> map = new HashMap<>();
            map.put("taskId", node.getId());
            map.put("markerId", node.getOwner());
            map.put("questions", JSON.toJSONString(node.getQuestions()));
            map.put("time", DateDisposeUtils.parseDateToStr(DateDisposeUtils.YYYY_MM_DD_HH_MM_SS, new Date(node.getTime())));
            list.add(map);
            node = node.next;
        }
        return list;
    }

    private static class LockNode {

        private LockNode previous;

        private LockNode next;

        private Object id;

        private Set<Long> questions;

        private Object owner;

        private long time;

        LockNode(Object id, Object owner, Set<Long> questions) {
            this.id = id;
            this.owner = owner;
            this.questions = questions;
            this.time = System.currentTimeMillis();
        }

        private boolean hasNext() {
            return this.next != null;
        }

        private boolean isId(Object id) {
            return this.id.equals(id);
        }

        private boolean isEqualQuestions(Set<Long> questions) {
            return CollectionUtils.isEqualCollection(this.questions, questions);
        }

        private boolean isOwner(Object owner) {
            return this.owner.equals(owner);
        }

        private boolean containsQuestionId(Long questionId) {
            return this.questions.contains(questionId);
        }

        private void remove() {
            this.previous.next = this.next;
            if (this.next != null) {
                this.next.previous = this.previous;
            }
        }

        private void append(Object id, Object owner, Set<Long> questions) {
            LockNode node = new LockNode(id, owner, questions);
            this.next = node;
            node.previous = this;
        }

        public Object getId() {
            return id;
        }

        public Object getOwner() {
            return owner;
        }

        public long getTime() {
            return time;
        }

        public Set<Long> getQuestions() {
            return questions;
        }
    }

}
